comment~    BINECHO.ASM for Win32
Reads hex text from the command line or from a file, converts it to binary
data, and sends the result to standard output. The hex text may include
ordinary text enclosed in single or double quotation marks.

To convert a file (such as a hex dump that has been manually edited)
to binary, just use

binecho <txtfile >outputfile
or
type txtfile | binecho > outputfile.

If StdIn has not been redirected as in those examples, Binecho reads text
from its command tail. E.g.
binecho 3C 'abc' 3E
produces
<abc>

Any character outside quotation marks, other than a hex digit, is
interpreted as a space.
These strings of input are equivalent:
770D0A
'w'0D0A
"w" d a.

Sample build file for binecho.asm:
@echo off
if exist binecho.exe del binecho.exe
\masm32\bin\ml /c /coff binecho.asm
if errorlevel 1 goto end
\masm32\bin\Link /SUBSYSTEM:CONSOLE binecho.obj
:end
dir binecho.exe
~

.386
.model flat, stdcall
option casemap:none

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
includelib \masm32\lib\kernel32.lib
cap_AL equ db 60,97,114,6,60,122,119,2,36,223  ;8086 machine code
ReadBuff_ = 512      ;buffer sizes
WriteBuff_ = 512

.data
hRead        dd ?    ;file handle
ReadBuff     dd ?    ;address of a buffer in the stack
amtReadBuff  dd 0    ;number of bytes of data remaining in the buffer
ptrReadBuff  dd ?    ;address of the next byte to get
hWrite       dd ?    ;file handle
WriteBuff    dd ?    ;address of a buffer in the stack
amtWriteBuff dd WriteBuff_    ;number of bytes of space remaining in the buffer
ptrWriteBuff dd ?    ;address to which to put the next byte

ByteHandler dd FirstDigit
ByteGetter  dd ReadCmdLine
EOFHandler  dd AnyRet

ByteGetters dd ReadCmdLine   ;FILE_TYPE_UNKNOWN = 0
            dd GetByte       ;FILE_TYPE_DISK = 1
            dd ReadCmdLine   ;FILE_TYPE_CHAR = 2
            dd GetByte       ;FILE_TYPE_PIPE = 3
cmd_tail dd ?

Nibble db ?
QuoteChar db 0

.code
Start:
    sub   esp,ReadBuff_
    mov   ReadBuff,esp
    sub   esp,WriteBuff_
    mov   WriteBuff,esp
    mov   ptrWriteBuff,esp
    invoke GetCommandLine
    xchg eax,edi
    mov  al,'"'
    scasb
    jne  short noquote
    or   ecx,-1
    repne scasb
    jmp  short got_cmd_tail
noquote:
    mov  al,[edi]
    test al,al
    jz   short got_cmd_tail
    inc  edi
    cmp  al," "
    je   short got_cmd_tail
    cmp  al,9
    jne  short noquote
got_cmd_tail:
    mov  cmd_tail, edi
    invoke GetStdHandle,-11   ; -11 = STD_OUTPUT_HANDLE
    mov  hWrite,eax
    invoke GetStdHandle,-10   ; -10 = STD_INPUT_HANDLE
    mov  hRead,eax
    invoke GetFileType,eax   ;see if standard input has been redirected
    test eax,-4
    jnz  short unknown_std_in
    mov  ecx,ByteGetters[4*eax]
    mov  ByteGetter,ecx
unknown_std_in:
MainLoop:
    call ByteGetter
    js   short @F
    call ByteHandler
    jmp  MainLoop
@@: call EOFHandler
    call FlushWrite
    invoke ExitProcess,0

GetByte:  ;gets a from the read handle into AL, or returns SF if end-of-file.
    mov esi,offset amtReadBuff
    dec dword ptr[esi]
    jns short @F   ;if read buffer is empty, read more of the input
    invoke ReadFile,hRead,ReadBuff,ReadBuff_,esi,0
    mov eax,[esi]
    dec eax
    js  short AnyRet
    mov [esi],eax
    mov eax,ReadBuff
    mov ptrReadBuff,eax
@@: mov esi,ptrReadBuff
    lodsb
    mov ptrReadBuff,esi
AnyRet:
    ret

PutNibble:
    mov al,Nibble
PutByte:
    mov edi,ptrWriteBuff
    stosb
    mov ptrWriteBuff,edi
    dec amtWriteBuff
    jnz short AnyRet
FlushWrite:
    mov eax,ptrWriteBuff
    mov edx,WriteBuff
    sub eax,edx
    mov ptrWriteBuff,edx
    invoke WriteFile,hWrite,edx,eax,addr amtWriteBuff,0
    mov amtWriteBuff,WriteBuff_
    ret

InQuotes:
    cmp al,QuoteChar
    jne short PutByte
    mov QuoteChar,0
    jmp short ByteDone

FirstDigit:
    call TestFirstQuote
    je   StartQuote
    call FromHex
    jnz  short @F
    mov  Nibble,al
    mov  ByteHandler,SecondDigit
    mov  EOFHandler,PutNibble
@@: ret

HangingDigit:
    call PutNibble
    mov  al,0
StartQuote:
    mov  QuoteChar,al
    mov  ByteHandler,InQuotes
    mov  EOFHandler,AnyRet
    ret

SecondDigit:
    call TestFirstQuote
    je   HangingDigit
    call FromHex
    jnz  OneDigit
    mov  cl,4
    shl  al,cl
    or   al,Nibble
    rol  al,cl
SecondDigitDone:
    call PutByte
ByteDone:
    mov ByteHandler,FirstDigit
    mov EOFHandler,offset AnyRet
    ret
OneDigit:
    mov al,Nibble
    jmp short SecondDigitDone

FromHex:     ;convert hex character al to a hex digit, return NZ if it isn't
    cap_AL
    cmp  al,"A"
    jb   short @F
    sub  al,"A"-1-"9"
@@: sub  al,"0"
    test al,0F0h
    ret

TestFirstQuote:
    cmp  al,"'"
    je   short @F
    cmp  al,'"'
    jne  short Ret2
@@: mov  QuoteChar,al
Ret2:
    ret

ReadCmdLine:
    mov  ecx,cmd_tail
    mov  al,[ecx]
    inc  ecx
    mov  cmd_tail,ecx
    test al,al
    setz ah
    neg  ah     ;ror ah,1 does not set the sign flag
    ret

end Start

debugging routines:

AX2hexEAX:
    mov  dl,ah
    call AL2hexAX
    ror  eax,16
    mov  al,dl
AL2hexAX:
    mov  ah,al
    shr  ah,4
    call Nibble2hexAL
    xchg al,ah
Nibble2hexAL:
    and  al,0Fh
    cmp  al,0Ah
    sbb  al,69h
    das
    ret

dump_eax:
    push 0A0Dh
    mov  ecx,eax
    call AX2hexEAX  ;saves ecx
    push eax
    xchg eax,ecx
    shr  eax,10h
    call AX2hexEAX
    push eax
    mov  eax,esp
    push edx
    mov  edx,esp
    invoke WriteFile,hWrite,eax,10,edx,0
    add  esp,16
    ret
